/*
 * xrdesktop
 * Copyright 2018 Collabora Ltd.
 * Author: Lubosz Sarnecki <lubosz.sarnecki@collabora.com>
 * SPDX-License-Identifier: MIT
 */

#ifndef XRD_H
#define XRD_H

#include <glib.h>

#define XRD_INSIDE

#include "xrd-version.h"
#include "xrd-button.h"
#include "xrd-client.h"
#include "xrd-desktop-cursor.h"
#include "xrd-container.h"
#include "xrd-input-synth.h"
#include "xrd-math.h"
#include "xrd-overlay-client.h"
#include "xrd-overlay-desktop-cursor.h"
#include "xrd-overlay-pointer.h"
#include "xrd-overlay-pointer-tip.h"
#include "xrd-overlay-window.h"
#include "xrd-render-lock.h"
#include "xrd-scene-background.h"
#include "xrd-scene-client.h"
#include "xrd-scene-desktop-cursor.h"
#include "xrd-scene-model.h"
#include "xrd-scene-object.h"
#include "xrd-scene-pointer.h"
#include "xrd-scene-pointer-tip.h"
#include "xrd-scene-renderer.h"
#include "xrd-scene-selection.h"
#include "xrd-scene-window.h"
#include "xrd-settings.h"
#include "xrd-shake-compensator.h"
#include "xrd-window.h"
#include "xrd-window-manager.h"
#include "xrd-enums.h"
#include "xrd-types.h"

#undef XRD_INSIDE

#endif /* XRD_H */
