/*
 * xrdesktop
 * Copyright 2020 Collabora Ltd.
 * Author: Lubosz Sarnecki <lubosz.sarnecki@collabora.com>
 * SPDX-License-Identifier: MIT
 */

#ifndef XRD_SCENE_CLIENT_PRIVATE_H_
#define XRD_SCENE_CLIENT_PRIVATE_H_

#include "xrd-scene-client.h"

G_BEGIN_DECLS

XrdSceneClient *
xrd_scene_client_new (GxrContext *context);

G_END_DECLS

#endif /* XRD_SCENE_CLIENT_PRIVATE_H_ */
